from littlelearn.GradientReflector.GradientReflector import GradientReflector, Node
import jax
import jax.numpy as jnp
from jax import device_put
from typing import Literal
import random
Global_engine_grad = GradientReflector()
float32 = jnp.float32
float16 = jnp.float16
float64 = jnp.float64
int32   = jnp.int32
int64   = jnp.int64
int16  = jnp.int16
bool    = jnp.bool_


def _get_device(device: Literal["cpu", "gpu"]):
    if device == "cpu":
        return jax.devices("cpu")[0]
    elif device == "gpu":
        gpus = jax.devices("gpu")
        if len(gpus) == 0:
            raise RuntimeError("No GPU available for JAX.")
        return gpus[0]
    else:
        raise ValueError(f"Unknown device: {device}")


class Tensor:
    def __init__(self, data, dtype=float32,
                 device: Literal["cpu", "gpu"] = "cpu",
                 requires_grad=False,_node : Node = None ):
        self.requires_grad = requires_grad
        self.is_param = False 

        self.dtype = dtype
        self.device = device
        self.node = _node

        if not isinstance(data, jnp.ndarray):
            arr = jnp.asarray(data, dtype=self.dtype)
        else:
            arr = data.astype(self.dtype)

        self.tensor = device_put(arr, _get_device(device))

        if self.requires_grad and self.node is None:
            self.node = Node(self.tensor)


            
    def to(self, device: Literal["cpu", "gpu"]):
        self.device = device
        self.tensor = device_put(self.tensor, _get_device(device))
        return self

    def __repr__(self):
        return (
            f"Tensor(shape={self.tensor.shape}, dtype={self.dtype}, "
            f"device={self.device}, requires_grad={self.requires_grad})\n"
            f"{self.tensor}"
        )

    @property
    def shape(self):
        return self.tensor.shape

    @property
    def grad(self):
        return None if self.node is None else self.node.grad

    @property
    def ndim(self):
        return self.tensor.ndim

    @property
    def size(self):
        return self.tensor.size
    
    def __add__ (self,other) : 
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.add(self,other)
    
    def __sub__(self,other)  :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.subtract(self,other)
    
    def pow(self,factor) :
        return Global_engine_grad.pow(self,factor=factor)
    
    def __mul__ (self,other) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.multiple(self,other)
    
    def __truediv__ (self,other) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.divide(self,other)
    
    def __pow__(self,factor) :
        return Global_engine_grad.pow(self,factor=factor)
    
    def __neg__(self) :
        return self *-1 
    
    def __radd__ (self,other) : 
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.add(other,self)

    def __rsub__ (self,other) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.subtract(other,self)
    
    def __rtruediv__ (self,other) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.divide(other,self)

    def __rmul__ (self,other) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.multiple(other,self)
    
    def matmul (self,other,tranpose_a=False,transpose_b = False) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.matmul(self,other,transpose_a=tranpose_a,transpose_b=transpose_b)
    
    def sum(self,axis=None,keepdims=False) :
        return Global_engine_grad.sum(self,axis=axis,keepdims=keepdims)
    
    def sumprod (self,other) :
        other = other if isinstance(other,Tensor) else Tensor(other,dtype=self.dtype,
                                                              requires_grad=self.requires_grad,
                                                              device=self.device)
        return Global_engine_grad.sumproduct(self,other)
    
    def mean(self,axis=None,keepdims=False) :
        return Global_engine_grad.mean(self,axis=axis,keepdims=keepdims)
    
    def var (self,axis=None,keepdims=False) :
        return Global_engine_grad.var(self,axis=axis,keepdims=keepdims)

    def std (self,axis=None,keepdims=False,epsilon=1e-6) :
        return Global_engine_grad.std(self,axis=axis,keepdims=keepdims,epsilon=epsilon)

    def exp(self) :
        return Global_engine_grad.exp(self)

    def log(self) :
        return Global_engine_grad.log(self)

    def log2(self) :
        return Global_engine_grad.log2(self)

    def log10(self) :
        return Global_engine_grad.log10(self) 
    
    def log1p(self) :
        return Global_engine_grad.log1p(self)
    
    def reshape(self,newshape : tuple) :
        return Global_engine_grad.reshape(self,newshape=newshape)
    
    def unsquezze (self,axis = None) :
        return Global_engine_grad.unsqueeze(self,axis=axis)
    
    def expand_dims (self,axis=None) :
        return Global_engine_grad.expand_dims(self,axis=axis)
    
    def squezze (self,axis=None) :
        return Global_engine_grad.squeeze(self,axis=axis)
    
    def transpose(self,shape : tuple) :
        return Global_engine_grad.transpose(self,axes=shape)
    
    def sqrt(self) :
        return Global_engine_grad.sqrt(self)
    
    def clip(self,min,max) :
        return Global_engine_grad.clip(self,min_value=min,max_value=max)
    
    def max (self,axis=None,keepdims=False) :
        return Global_engine_grad.max(self,axis=axis,keepdims=keepdims)
    
    def min (self,axis=None,keepdims=False) :
        return Global_engine_grad.min(self,axis=axis,keepdims=keepdims)
    
    
    def flatten (self) :
        return Global_engine_grad.flatten(self)
    
    def where (self,condition,then_x,then_y) :
        return Global_engine_grad.where(condition,then_x,then_y)
    

    def argmax (self,axis = None) :
        return Tensor(jnp.argmax(self.tensor,axis=axis) )
    
    def argmin (self,axis=None) :
        return Tensor(jnp.argmin(self.tensor,axis=axis))
    
    def __getitem__ (self,idx) :
        return Global_engine_grad.getitem(self,idx)
    
    def abs (self) :
        return Global_engine_grad.abs(self)
    
    def pad (self,pad_with,mode='constant',constant_value=0) :
        return Global_engine_grad.pad(self,pad_width=pad_with,mode=mode,
                                      constant_values=constant_value)
    
    def broadcast_to (self,shape: tuple) :
        return Global_engine_grad.broadcast_to(self,shape=shape)
    
    def erf (self) :
        return Global_engine_grad.erf(self)
    
    def sinh (self) :
        return Global_engine_grad.sinh(self)
    
    def cosh (self) :
        return Global_engine_grad.cosh(self)
    
    def tan (self) :
        return Global_engine_grad.tan(self)
    
    def floor (self) :
        return Global_engine_grad.floor(self)
    
    def ceil (self) :
        return Global_engine_grad.ceil(self)
    
    def rint (self) :
        return Global_engine_grad.rint(self)
    
    def reciprocal (self) :
        return Global_engine_grad.reciprocal(self)
    
    def sign (self) :
        return Global_engine_grad.sign(self)
    
    def identity (self) :
        return Global_engine_grad.identity(self)
    
    def concat(self,tensor_b,axis=None) :
        tensor_b = tensor_b if isinstance(tensor_b,Tensor) else Tensor(tensor_b,
                                                                       dtype=self.dtype,
                                                                       device=self.device,
                                                                       requires_grad=self.requires_grad)
        return Global_engine_grad.concat([self,tensor_b],axis=axis)
    
    def logsumexp (self,axis=None,keepdims=False) :
        return Global_engine_grad.logsumexp(self,axis=axis,keepdims=keepdims)
    
    def logaddexp (self,b) :
        b = b if isinstance(b,Tensor) else Tensor(b,dtype=self.dtype,
                                                  device=self.device,
                                                  requires_grad=self.requires_grad) 
        return Global_engine_grad.logaddexp(self,b)
    
    def sin(self) :
        return Global_engine_grad.sin(self)
    
    def cos (self) :
        return Global_engine_grad.cos(self)

    
    def backwardpass (self) :
        if self.node is not None : 
            self.node.backwardpass()
    
    def reset_grad (self) :
        if self.node is not None :
            self.node.reset_grad()
    
    def nonactive_grad (self) :
        self.node = None 
        self.requires_grad = False 
    
    def active_grad (self) :
        if self.node is None :
            self.node = Node(self.tensor,parents=())
        self.requires_grad = True


def randn(shape : tuple,dtype = float32,device : Literal["cpu","gpu"] = "cpu",requires_grad = False,
          max_random_seed = 10 ) -> Tensor :
    random_seed = random.randint(0,max_random_seed)
    data = jax.random.normal(jax.random.PRNGKey(random_seed),shape).astype(dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def rand (*args : int) -> Tensor :
    random_seed = random.randint(0,100)
    data = jax.random.uniform(jax.random.PRNGKey(random_seed),shape=args)
    return Tensor(data)

def uniform (low : float = -1.0,high : float = 1.0, shape : tuple = (), 
             dtype=float32,device : Literal["cpu","gpu"]="cpu", max_random_seed = 10,
             requires_grad = False) :
    random_keys = random.randint(0,max_random_seed) 
    data = jax.random.uniform(
        key=jax.random.PRNGKey(random_keys),shape=shape,dtype=dtype,
        minval=low,maxval=high)
    
    return Tensor(data,device=device,requires_grad=requires_grad)

def zeros(shape : tuple,dtype=float32,device : Literal["cpu","gpu"] = "cpu",requires_grad = False) -> Tensor :
    data = jnp.zeros(shape,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def ones(shape : tuple,dtype=float32,device : Literal["cpu","gpu"] = "cpu",requires_grad = False) -> Tensor :
    data = jnp.ones(shape,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def arange (start : int, end : int, step : int =1 , dtype = int32, device : Literal["cpu","gpu"] ="cpu", requires_grad = False) -> Tensor :
    data = jnp.arange(start,end,step,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def eye (n : int, dtype = float32, device : Literal["cpu","gpu"] ="cpu", requires_grad = False) -> Tensor :
    data = jnp.eye(n,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def arange_like (tensor : Tensor, start : int, end : int, step : int =1 , dtype = int32, device : Literal["cpu","gpu"] ="cpu", requires_grad = False) -> Tensor :
    data = jnp.arange(start,end,step,dtype=dtype)
    data = data.reshape(tensor.shape)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def eye_like (tensor : Tensor, dtype = float32, device : Literal["cpu","gpu"] ="cpu", requires_grad = False) -> Tensor :
    n = tensor.shape[0]
    data = jnp.eye(n,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def randn_like (tensor : Tensor,dtype = float32,device : Literal["cpu","gpu"] = "cpu",requires_grad = False) -> Tensor :
    data = jax.random.normal(jax.random.PRNGKey(random.randint(0,100)),
                             tensor.shape)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def rand_like (tensor : Tensor) -> Tensor :
    data = jax.random.uniform(jax.random.PRNGKey(random.randint(0,100)),shape=tensor.shape)
    return Tensor(data)

def zeros_like (tensor : Tensor,dtype=float32,device : Literal["cpu","gpu"] = "cpu",requires_grad = False) -> Tensor :
    data = jnp.zeros(tensor.shape,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def ones_like (tensor : Tensor,dtype=float32,device : Literal["cpu","gpu"] = "cpu",requires_grad = False) -> Tensor :
    data = jnp.ones(tensor.shape,dtype=dtype)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)

def tril (tensor : Tensor,diagonal=0,dtype=float32,device : Literal["cpu","gpu"]="cpu",
          requires_grad = False) -> Tensor:
    data = jnp.tril(tensor.tensor,k=diagonal)
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)


def binomial (n :float , p : float , shape : tuple ,max_random_keys = 10,
              device = "cpu") :
    max_random = random.randint(0,max_random_keys)
    data = jax.random.binomial(
        key=jax.random.PRNGKey(max_random),n=n,p=p,shape=shape
    )
    return Tensor(data=data,device=device)

def top_k (logits : Tensor,top_k = 1) :
    logits = jax.lax.top_k(logits.tensor,top_k)
    return Tensor(logits)


def categorical (logits : Tensor,max_random_seed = 10,axis=-1) :
    keys = jax.random.PRNGKey(random.randint(0,max_random_seed))
    categori = jax.random.categorical(key=keys,logits=logits.tensor,axis=axis)
    return Tensor(categori)

def orthogonal (size : int = 1,dtype=float32,device : Literal["cpu","gpu"]="cpu", max_random_seed = 10,
             requires_grad = False) :
    keys= jax.random.PRNGKey(random.randint(0,max_random_seed))
    logits = jax.random.normal(keys,shape=(size,size))
    Q,R = jnp.linalg.qr(logits)

    data = jnp.sign(jnp.diag(R))
    data = Q * data
    return Tensor(data,dtype=dtype,device=device,requires_grad=requires_grad)


def normal(mean :float = 0.0, std = 1.0,shape : tuple = ()
           ,dtype=float32,device : Literal["cpu","gpu"]="cpu", max_random_seed = 10,
             requires_grad = False) :
    keys = jax.random.PRNGKey(random.randint(0,max_random_seed))
    data = jax.random.normal(keys,shape=shape)
    data = mean + std * data 
    return Tensor(data=data,dtype=dtype,device=device,
                  requires_grad=requires_grad) 


from littlelearn.DeepLearning.activations import Activations as ac
from littlelearn.tensor import Tensor,float32,rand,uniform,zeros,ones
import littlelearn as ll
from typing import Literal
import jax.numpy as jnp 
import numpy as np 
import math 

class Parameter(Tensor):
    def __init__(self, tensor, dtype=float32):
        
        if isinstance(tensor, Tensor):
            super().__init__(
                data=tensor.tensor,
                dtype=dtype,
                requires_grad=False
              
            )
        else:

            super().__init__(
                data=tensor,
                dtype=dtype,
                requires_grad=False
            )
        
        self.is_param = True
    
    def float_16_param (self) :
        self.tensor = jnp.asarray(self.tensor,dtype=jnp.float16)

class Component :

    """
        Component
        ---------------
        Component is abstract layers template for make custom model or 
        spesial model without need to make some function to updating weight.

        How to use:
        ----------------
            ```
                class MLP(Component):
                    def __init__ (self,num_class):
                        super().__init__()
                        self.layers1 = Dense(32,'relu')
                        self.layers2 = Dense(64,'relu')
                        self.final_out = Dense(num_class)
                    
                    def __call__(self,**kwargs):
    
                        x = self.layers1(x)
                        x = self.layers2(x)
                        x = self.layers3(x)
            ```
        
        Author:
        ---------
        Candra Alpin Gunawan 

        Warning:
        ------------------
        you have call  Component or model before get_weight  
    """

    def __init__(self):
        pass 

    def load_state(self,model_path="model.npz"):
        
        data = np.load(model_path)
        for i, param in enumerate(self.parameter()):
            key = f"param{i}"
            if key in data:
                param.tensor = jnp.array(data[key])
        print(f"model load from {model_path}")
    
    def save_state(self, model_path="model.npz"):
        param_dict = {f"param{i}": np.array(p.tensor) for i, p in enumerate(self.parameter())}
        np.savez(model_path, **param_dict)
        print(f"model saved at {model_path}")
    
    def parameter(self) :
        param = list()

        for obj in self.__dict__.values():

            if isinstance(obj,Parameter):
                param.append(obj)
            elif isinstance(obj,Component):
                weight = obj.parameter()
                for w in weight :
                    param.append(w) 
            elif isinstance(obj,list) :
                for ob in obj :
                    if isinstance(ob,Component) :
                        weight = ob.parameter()
                        for w in weight :
                            param.append(w) 


        return param 
    
    def to (self,device : str) :
        print(f"{self.__class__.__name__} to {device}")
        for obj in self.parameter() :

            if isinstance(obj,Parameter) :
                obj.to(device)
            

    
    def train(self,state:bool=True) :
        if state :
            
            for obj in self.parameter() :
                if isinstance(obj,Parameter) :
                    obj.active_grad()
            print(f"name : {self.__class__.__name__ } status : active || requires_grad : {obj.requires_grad}")
        else :
            print("all object still inference Mode")

    def inference (self,state : bool = True) :
        if state :
            
            for obj in self.parameter():
                if isinstance(obj,Parameter) :
                    obj.nonactive_grad()
            print(f"name : {self.__class__.__name__ } status : non active || requires_grad : {obj.requires_grad}")
        else :
            print("all object still train Mode")
    
    def set_param_to_float16 (self) :
        for param in self.parameter():
            if isinstance(param,Parameter) :
                param.float_16_param()

    def forwardpass (self,*args,**kwargs) :
        raise NotImplementedError("Define forwardpass() in this class ") 


    def __call__(self, *args, **kwargs):
        return self.forwardpass(*args,**kwargs) 
    

 
class Linear (Component) :

    def __init__ (self,input_dim,output_dim,add_bias=True,dtype=float32) :
        super().__init__()
        self.dtype = dtype
        glorot = math.sqrt(6/(input_dim + output_dim))
        w = uniform(low=-glorot,high=glorot,shape=(input_dim,output_dim),max_random_seed=1000)
        self.weight = Parameter(w,dtype=self.dtype)
        if add_bias :
            bias = zeros((1,output_dim),dtype=dtype)
            self.bias = Parameter(bias,dtype=dtype)
        else :
            self.bias = None 
    
    def forwardpass(self,x):
        if not isinstance(x,Tensor) :
            raise ValueError("input must be Tensor ")
        if self.bias is not None :
            return ll.matmul(x,self.weight) + self.bias
        
        return ll.matmul(x,self.weight)
    
class Embedding (Component) :
    def __init__(self,vocab_size : int , embedding_dim : int) :
        super().__init__()
        self.weight = Parameter(tensor=rand(vocab_size,embedding_dim))
    
    def forwardpass(self,x):
        if isinstance(x,Tensor) :
            x = x.tensor
      
        return self.weight[x]

class GlobalAveragePooling1D (Component) :
    def __init__(self,axis=1,keepdims=False):
        super().__init__() 
        self.axis = axis 
        self.keepdims = keepdims
    
    def forwardpass(self, x):
        if not isinstance(x,Tensor) :
            raise ValueError("input must be Tensor")
        return ll.mean(x,axis=self.axis,keepdims=self.keepdims)


class Attention (Component) :
    def __init__ (self,embed_dim : int,add_bias : bool = False, return_attention = False ,use_causal_mask = False):
        super().__init__()
        std = math.sqrt(6/(embed_dim * 2 ))
        self.wq = Parameter(tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100))
        self.wk = Parameter(tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100))
        self.wv = Parameter(tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100))
        if add_bias :
            self.bq = Parameter(tensor=zeros(shape=(1,embed_dim)))
            self.bk = Parameter(tensor=zeros(shape=(1,embed_dim)))
            self.bv = Parameter(tensor=zeros(shape=(1,embed_dim)))
        else :
            self.bq = None 
            self.bv = None 
            self.bk = None 
        self.is_causal_mask = use_causal_mask
        self.return_attention = return_attention
        self.__add_bias = add_bias
    
    def __created_causal (self,size,device) :
        return 1 - ll.tril(ones(shape=(size,size)),diagonal=0,device=device)
    
    def __scaled_dot_product(self,Q,K,V,causal_mask = None):
        dim_k = ll.to_tensor(K.shape[-1],device=Q.device)
        logits = ll.matmul(Q,K,transpose_b=True) / ll.sqrt(dim_k)
        if causal_mask is not None :
            logits = logits + causal_mask
        score = ll.DeepLearning.activations.softmax(
            x=logits,axis=-1,keepdims=True,use_crossentropy=False
        )
        out = ll.matmul(score,V)
        if self.return_attention :
            return out,score
        else :
            return out 
    
    def forwardpass(self,query,keys,values) :
        if self.__add_bias  :
            q = ll.matmul(query,self.wq) + self.bq
            k = ll.matmul(keys,self.wk) + self.bk 
            v = ll.matmul(values,self.wv) + self.bv 
        
        else :
            q = ll.matmul(query,self.wq)
            k = ll.matmul(keys,self.wk)
            v = ll.matmul(values,self.wv)
        
        if self.is_causal_mask:
        
            mask = self.__created_causal(q.shape[1],device=q.device) * -1e9

            return self.__scaled_dot_product(q,k,v,mask)
        else :
            return self.__scaled_dot_product(q,k,v,causal_mask=None)



class RMSNorm (Component) :
    def __init__ (self,embed_dim : int,epsilon = 1e-6) :
        super().__init__()
        self.weight = Parameter(tensor=ones(shape=(1,embed_dim)))
        self.eps = epsilon
    
    def forwardpass(self,x) :
        rms = ll.sqrt(ll.mean(a=x**2,axis=-1,keepdims=True) + self.eps)
        rms_x = x / rms 
        return rms_x * self.weight

class Dropout(Component) :
    def __init__(self,rate :float = 0.1) :
        super().__init__()
        self.rate = rate 
    
    def forwardpass(self,x) :
        rate = 1 - self.rate
        m = ll.binomial(1,rate,shape=x.shape,device=x.device)
        x = m * x / rate 
        return x 

class LayerNorm (Component) :
    def __init__ (self,embed_dim : int, epsilon=1e-6) :
        super().__init__()
        self.embed_dim = embed_dim
        self.eps = epsilon
        self.w_gamma = Parameter(tensor=ones(shape=(1,self.embed_dim)))
        self.w_beta = Parameter(tensor=zeros(shape=(1,self.embed_dim)))
    
    def forwardpass(self, x):
        mean = ll.mean(x,axis=-1,keepdims=True)
        var = ll.var(x,axis=-1,keepdims=True)
        shifted1 = x - mean 
        shifted2 = ll.sqrt(var + self.eps)
        return shifted1 / shifted2 * self.w_gamma + self.w_beta

class BatchNorm (Component) :
    def __init__(self,embed_dim : int , epsilon = 1e-6) :
        super().__init__()
        self.embed_dim = embed_dim 
        self.eps = epsilon
        self.w_gamma = Parameter(tensor=ones(shape=(1,self.embed_dim)))
        self.w_beta = Parameter(tensor=zeros(shape=(1,self.embed_dim)))
    
    def forwardpass (self,x) :
        mean = ll.mean(x,axis=0,keepdims=True)
        var = ll.var(x,axis=0,keepdims=True)
        shifted1 = x - mean 
        shifted2 = ll.sqrt(var + self.eps)
        return shifted1 / shifted2 * self.w_gamma + self.w_beta


class SimpleRNN (Component) :
    def __init__ (self,embed_dim : int,return_sequence = False ) :
        super().__init__()
        self.embed_dim = embed_dim
        self.return_sequence = return_sequence 
        std = math.sqrt(6 / (embed_dim * 2))
        self.weight_s = Parameter(tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=0))
        self.weight_h = Parameter(tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=0)) 
        self.bias = Parameter(tensor=zeros(shape=(1,embed_dim))) 
        self.hidden_state = None 
    
    def forwardpass(self,x) :
        B,T,D = x.shape 
        self.hidden_state=zeros(shape=(B,D),requires_grad=False,device=x.device)
        output = list()
        for t in range(T) :
            x_t = x[:,t,:]
            h_new = (ll.matmul(x_t,self.weight_s) + ll.matmul(self.hidden_state,self.weight_h) + self.bias)
            h_new = ll.DeepLearning.activations.tanh(h_new)
            output.append(h_new.reshape((B,1,D)))
            self.hidden_state = h_new
        
        
        output =  ll.concat(output,axis=1)

        if self.return_sequence :
            return output
        else :
            return output[:,-1,:] 
 

class MultiHeadAttention (Component) :
    def __init__ (self,embed_dim : int ,num_head : int =None,add_bias=False,return_attention=False,
                  use_causal_mask = False) :
        super().__init__()
        std = math.sqrt(6/(embed_dim * 2))
        self.return_attention = return_attention
        self.use_causal_mask = use_causal_mask
        self.add_bias = add_bias
        self.wq = Parameter(
            tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100)
        )
        self.wk = Parameter(
             tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100)
        )
        self.wv = Parameter(
             tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100)
        )

        self.wo = Parameter(
            tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=100)
        )

        if add_bias :
            self.bq = Parameter(
                tensor=zeros(shape=(1,embed_dim))
            ) 
            self.bk = Parameter(
                tensor=zeros(shape=(1,embed_dim))
            )
            self.bv = Parameter(
                tensor=zeros(shape=(1,embed_dim))
            )
            self.bo = Parameter(
                tensor=zeros(shape=(1,embed_dim))
            )
        else :
            self.bq = None 
            self.bk = None 
            self.bv = None 
        
        if num_head is None :
            if embed_dim >=64 : 
                num_head = embed_dim // 64
            else :
                num_head = 2 

        self.num_head = num_head
        self.dim_k = embed_dim // num_head

    
    def __create_causal_mask (self,size,device) :
        return 1 - ll.tril(ones(shape=(size,size)),diagonal=0,device=device)
    
        
    
    def __split_head(self,X) :
        B,S,_  = X.shape
        X = ll.reshape(X,(B,S,self.num_head,self.dim_k))
        X = ll.transpose(X,(B,self.num_head,S,self.dim_k))
        return X
    
    def __scaled_dot_product (self,Q,K,V,causal_mask=None):
        dim_k = ll.to_tensor(K.shape[-1],device=Q.device)
        score = ll.matmul(Q,K,transpose_b=True) / ll.sqrt(dim_k)
        if causal_mask is not None :
            score = score + causal_mask
        
        attention = ll.DeepLearning.activations.softmax(
            x=score,axis=-1,keepdims=True,use_crossentropy=False
        )

        output = ll.matmul(attention,V)

        return output,attention

    def forwardpass (self,query,keys,values) :
        B,S,D = query.shape
        if self.add_bias :
            q = self.__split_head(ll.matmul(query,self.wq) + self.bq)
            k = self.__split_head(ll.matmul(keys,self.wk) + self.bk)
            v = self.__split_head(ll.matmul(values,self.wv) + self.bv)
        else :
            q = self.__split_head(ll.matmul(query,self.wq))
            k = self.__split_head(ll.matmul(keys,self.wk))
            v = self.__split_head(ll.matmul(values,self.wv))
        
        if self.use_causal_mask :
            mask = self.__create_causal_mask(size=S,device=q.device) * -1e9
            output,attn = self.__scaled_dot_product(
                Q=q,K=k,V=v,causal_mask=mask
            )
        else :
            output,attn = self.__scaled_dot_product(
                Q=q,K=k,V=v,causal_mask=None
            )

        output = ll.transpose(output,new_shape=(B,S,self.num_head,self.dim_k))
        output = ll.reshape(output,new_shape=(B,S,D))

        if self.add_bias :
            output = ll.matmul(output,self.wo) + self.bo
        
        else :
            output = ll.matmul(output,self.wo)

        if self.return_attention :
            return output,attn
        
        else :
            return output
        
class LSTM (Component) :
    def __init__ (self,embed_dim : int, return_sequence = True,return_state = False) :
        super().__init__()
        std = math.sqrt(float(embed_dim))
        self.w_input = Parameter(
            tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=20)
        )
        self.w_forgot = Parameter(
            tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=20)
        )
        self.w_gate = Parameter(
            tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=20)
        )
        self.w_output = Parameter(
            tensor=uniform(low=-std,high=std,shape=(embed_dim,embed_dim),max_random_seed=20)
        )
        self.wh_in = Parameter(
            tensor=ll.orthogonal(size=embed_dim)
        )
        self.wh_fo = Parameter(
            tensor=ll.orthogonal(size=embed_dim)
        )
        self.wh_ga = Parameter(
            tensor=ll.orthogonal(size=embed_dim)
        )
        self.wh_o = Parameter(
            tensor=ll.orthogonal(size=embed_dim)
        )
        self.return_sequence = return_sequence
        self.return_state = return_state
        self.hidden_state = None 
        self.cell_state = None 
    
    def forwardpass (self,x,state = None) :
        B,T,D = x.shape
        if state is not None :
            self.hidden_state = state[0]
            self.cell_state = state[1]
        else :
            self.hidden_state = zeros(shape=(B,D),requires_grad=x.requires_grad,device=x.device)
            self.cell_state = zeros(shape=(B,D),requires_grad=x.requires_grad,device=x.device)
        output = list()
        for t in range(T) :
            x_t = x[:,t,:]
            self.cell_state = ll.GradientReflector.lstm_cell(
                x_t=x_t,h_prev=self.hidden_state,c_prev=self.cell_state,
                w_gate=(self.w_input,self.w_forgot,self.w_forgot),
                wh_gate=(self.wh_in,self.wh_fo,self.wh_ga)
            ) 
            self.hidden_state = ll.GradientReflector.lstm_hidden(
                x_t=x_t,h_prev=self.hidden_state,cell_state=self.cell_state,
                w_output=self.w_output,wh_output=self.w_output
            )
        
            output.append(self.hidden_state.reshape((B,1,D)))
        output = ll.concat(output,axis=1)
        if self.return_sequence :
            if self.return_state :
                return output,self.hidden_state,self.cell_state
            else :
                return output
        
        else :
            if self.return_state :
                return output[:,-1,:],self.hidden_state,self.cell_state
            else :
                return output[:,-1,:]
            
           

class Sequential (Component) :
    def __init__ (self,component : list) :
        super().__init__()
        self.__component = component
    

    def parameter(self):
        param = []
        for obj in self.__component :
            if isinstance(obj,Component) :
                weight = obj.parameter()
                for w in weight :
                    param.append(w)
        
        return param 

    def forwardpass(self,x) :
        for component in self.__component :
            if isinstance(component,Component) :
                x = component(x)
        
        return x 

class LCMBlock (Component) :
    def __init__(self,embed_dim: int,drop_rate :float = 0.1 ) :
        super().__init__()
        self.step1 = Linear(embed_dim,embed_dim)
        self.step2 = Linear(embed_dim,embed_dim)
        self.norm = RMSNorm(embed_dim)
        self.dropout = Dropout(drop_rate)
        self.magnitude_layers = Linear(embed_dim,embed_dim)

    def forwardpass(self,x) :
        prenorm = self.norm(x)
        step1 = self.step1(prenorm)
        step1 = ll.DeepLearning.activations.gelu(step1)
        step2 = self.step2(prenorm)
        step2 = ll.DeepLearning.activations.gelu(step2)
        laten = step1 + step2 
        laten = self.dropout(laten)
        laten = self.magnitude_layers(laten)
        laten = ll.DeepLearning.activations.tanh(laten)
        return x + laten 

class LCTBlock (Component) :
    def __init__ (self,embed_dim : int ,num_head : int,drop_rate : float = 0.1,
                  use_causal_mask = False) :
        super().__init__()
        self.attention = MultiHeadAttention(embed_dim=embed_dim,num_head=num_head,use_causal_mask=use_causal_mask)
        self.norm1 = RMSNorm(embed_dim)
        self.norm2 = RMSNorm(embed_dim)
        self.dropout = Dropout(drop_rate)
        self.lcm = LCMBlock(embed_dim,drop_rate)
    
    def forwardpass (self,x) :
        norm1 = self.norm1(x)
        attn = self.attention(norm1,norm1,norm1)
        attn = self.dropout(attn)
        x = x + attn 

        norm2 = self.norm2(x)
        lcm = self.lcm(norm2)
        
        return lcm 


class FeedForwardNetwork (Component) :
    def __init__(self,embed_dim : int ,ffn_dim : int):
        super().__init__()
        self.linear1 = Linear(embed_dim,ffn_dim)
        self.gelu = ll.DeepLearning.activations.Gelu()
        self.linear2 = Linear(ffn_dim,embed_dim) 
    
    def forwardpass(self,x) :
        x = self.linear1(x)
        x = self.gelu(x)
        x = self.linear2(x)
        return x 


class TransformersBlock (Component) :
    def __init__(self,embed_dim : int,ffn_dim : int,drop_rate : float = 0.1,type : Literal ['single','multi']= 'multi',
                 mode : Literal['Encoder','Decoder'] = 'Encoder') :
        super().__init__()
        if type == 'multi' :
            if mode == 'Encoder':
                self.attention = MultiHeadAttention(embed_dim=embed_dim)
            elif mode == 'Decoder':
                self.attention = MultiHeadAttention(embed_dim=embed_dim,use_causal_mask=True)
            else :
                raise ValueError(f"{type} not  supported")
        elif type == 'single':
            if mode == 'Encoder' :
                self.attention = Attention(embed_dim=embed_dim)
            elif mode == 'Decoder':
                self.attention = Attention(embed_dim=embed_dim,use_causal_mask=True)
            else :
                raise ValueError(f"{type} not  supported")
        else :
            raise ValueError(f"{type} not supported") 
        
        self.ffn = FeedForwardNetwork(embed_dim,ffn_dim)

        self.norm1 = RMSNorm(embed_dim)
        self.norm2 = RMSNorm(embed_dim)
        self.drop_out1 = Dropout(drop_rate)
        self.drop_out2 = Dropout(drop_rate)
    
    
    def forwardpass(self,x) :
        norm1 = self.norm1(x)
        attn = self.attention(norm1,norm1,norm1)
        attn = self.drop_out1(attn)
        x = x + attn 

        norm2 = self.norm2(x) 
        ffn = self.ffn(norm2)
        x = self.drop_out2(x)
        x = x + ffn 

        return x

class DiagonalSSM(Component):
    def __init__(self, embed_dim: int):
        super().__init__()

 
        d_state = embed_dim

        self.state_w = Parameter(
            tensor=ll.DeepLearning.activations.tanh(
                ll.randn(shape=(d_state,), max_random_seed=30)
            )
        )

        self.wb = Parameter(
            tensor=ll.randn(shape=(d_state, embed_dim), max_random_seed=30)
        )

        self.wc = Parameter(
            tensor=ll.randn(shape=(embed_dim, d_state), max_random_seed=30)
        )

        self.Bd = Parameter(
            tensor=ll.randn(shape=(embed_dim,))
        )

        self.wg = Parameter(
            tensor=ll.randn(shape=(embed_dim, d_state), max_random_seed=30)
        )

    def forwardpass(self, u):
        B, T, D = u.shape
        d_state = self.state_w.shape[0]

        x = zeros((B, d_state), device=u.device)
        output = []

        for t in range(T):
            u_t = u[:, t]

            
            gate = ll.DeepLearning.activations.sigmoid(
                ll.matmul(u_t, self.wg, transpose_b=True)
            ) 


            x_hat = self.state_w * x + ll.matmul(
                u_t, self.wb, transpose_b=True
            )


            x = gate * x_hat + (1.0 - gate) * x


            y = (
                ll.matmul(x, self.wc, transpose_b=True)
                + u_t * self.Bd
                + u_t
            )

            output.append(y.reshape((B, 1, D)))

        return ll.concat(output, axis=1)

class Conv2d (Component) :
    def __init__(self,input_chanel : int , output_chanel : int,kernel_2n : int =2 ,
                 stride : int = 1 ,padding=0,add_bias = False):
        super().__init__()
        std = math.sqrt(6/(input_chanel + output_chanel))
        self.weight = Parameter(
            tensor= uniform(
            low = -std,high=std,shape=(output_chanel,input_chanel,kernel_2n,kernel_2n),
            max_random_seed=100
            ) 
        ) 
        
        if add_bias :
            self.bias = Parameter(
                tensor=zeros(
                    shape=(1,output_chanel)
                )
            )
        else :
            self.bias = None 
        
        self.stride = stride
        self.padding = padding
    
    def forwardpass(self,x) :
        return ll.GradientReflector.conv2d(
            x = x,w=self.weight,b=self.bias,stride=self.stride,pad=self.padding
        )

class GlobalAveragePooling2d (Component) :
    def __init__ (self,axis=(-1,-2),keepdims=False) :
        super().__init__()
        if len(axis) != 2 :
            raise ValueError(f"{axis} error this layers just support 2d axis = (axis1,axis2)")
        self.axis = axis 
        self.keepdims = keepdims
    
    def forwardpass(self,x) :
        return ll.mean(x,axis=self.axis,keepdims=self.keepdims) 

class Maxpool2d (Component) :
    def __init__(self,kernel_2t=2,stride_t=2,padding=0) :
        super().__init__()
        self.kernel = (kernel_2t,kernel_2t)
        self.stride = (stride_t,stride_t)
        self.padding = padding
    
    def forwardpass(self,x) :
        return ll.GradientReflector.maxpool2d(
            x=x,kernel=self.kernel,stride=self.stride,
            padding=self.padding
        )



"""
# LittleLearn: Touch the Big World with Little Steps 🌱

LittleLearn is an original, experimental machine learning framework — inspired by
the simplicity of Keras and the flexibility of PyTorch — but designed from scratch
with its own architecture, philosophy, and backend engine.

## 🔖 Note:

Although inspired by Keras and PyTorch, **LittleLearn is an original framework** built from the ground up.
It aims to provide a new perspective on model building — flexible, introspective, and fully customizable.

Author: 
----------
Candra Alpin Gunawan 
"""
__name__ = "littlelearn"
__author__ = "Candra Alpin Gunawan"
__version__ = "1.0.7"
__license__ = "Apache 2.0"
__realese__ = "29-December-2025"
__email__ = "hinamatsuriairin@gmail.com"
__repo__ = "https://github.com/Airinchan818/LittleLearn"
__youtube__ = "https://youtube.com/@hinamatsuriairin4596?si=KrBtOhXoVYnbBlpY"

from . import preprocessing
from .tensor import *
from .GradientReflector import GradientReflector
from .GradientReflector import Node
from . import DeepLearning


def to_tensor (x,dtype = float32,device="cpu",requires_grad=True) :
    if isinstance (x,Tensor) :
        print("warning data has being Tensor, is just rebuild Tensor in the same Tensor object")
        x = x.tensor
    return Tensor(x,dtype=dtype,device=device,requires_grad=requires_grad)

def matmul (a,b,transpose_a=False,transpose_b=False) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    if not isinstance(b,Tensor) :
        b = Tensor(b)
    return a.matmul(b,tranpose_a=transpose_a,transpose_b=transpose_b)

def sum(a,axis=None,keepdims=False) :
    if not isinstance(a,Tensor) : 
        a = Tensor(a)
    
    return a.sum(axis=axis,keepdims=keepdims)

def mean(a,axis=None,keepdims=False) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.mean(axis=axis,keepdims=keepdims)

def log (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.log()

def exp(a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.exp()

def flatten (a) :
  if not isinstance(a,Tensor) :
      a = Tensor(a)
  
  return a.flatten()

def expand_dims (a,axis=0) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.expand_dims(axis=axis)

def squezze (a,axis=None) :
  if not isinstance(a,Tensor) :
      a = Tensor(a)
  
  return a.squezze(axis=axis)

def unsquezze (a,axis=None) :
    if not isinstance(a,Tensor) :
        a= Tensor(a)
    
    return a.unsquezze(axis=axis)

def sumprod (a,b) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    if not isinstance(b,Tensor) :
        b = Tensor(b)
    
    return a.sumprod(a,b)

def var(a,axis=None,keepdims=False) :
    if not isinstance(a,Tensor)  :
        a = Tensor(a)
    
    return a.var(axis=axis,keepdims=keepdims)

def std(a,axis=None,keepdims=False,epsilon=1e-5) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.std(axis=axis,keepdims=keepdims,epsilon=1e-5)

def log2 (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.log2()

def log10(a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.log10()

def log1p(a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.log1p()

def reshape(a,new_shape: tuple) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    return a.reshape(newshape=new_shape)

def transpose(a,new_shape:tuple) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.reshape(newshape=new_shape)

def sqrt (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.sqrt()

def clip(a,min_vals :float,max_vals:float) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.clip(min_vals,max_vals)

def max(a,axis=None,keepdims=False) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.max(axis=axis,keepdims=keepdims)

def min (a,axis=None,keepdims=False) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.min(axis=axis,keepdims=keepdims)

def argmax(a,axis=None) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.argmax(axis=axis)

def abs (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.abs()

def pad(a,pad_with,mode='constant',constant_value=0) :
    if not isinstance(a,Tensor) :
        a =  Tensor(a)
    
    return a.pad(pad_with=pad_with,mode=mode,constant_value=constant_value)

def broadcast_to (a,shape :tuple) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.broadcast_to(shape)

def erf (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.erf()

def sinh (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.sinh()

def consh (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.cosh()

def tan (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.tan()

def floor (a) : 
    if not isinstance(a,Tensor) : 
        a = Tensor(a)
    
    return a.floor()

def ceil (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.ceil()

def rint (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.rint()

def reciprocal (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.reciprocal()

def sign (a) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.sign()

def identity (a) :
    if not isinstance(a,Tensor) : 
        a = Tensor(a)
    
    return a.identity()

def concat (tensor_list : list,axis=0) :
    if not isinstance(tensor_list[0],Tensor) :
        for i in range(len(tensor_list)) :
            tensor_list[i] = Tensor(tensor_list[i])
    
    return GradientReflector.concat(tensor_list,axis=axis)

def logsumexp (a,axis = None,keepdims=False) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.logsumexp(axis=axis,keepdims=keepdims)

def logaddexp (a,b) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    
    return a.logaddexp(a,b)

def argmin (a,axis=None) :
    if not isinstance(a,Tensor) :
        a = Tensor(a)
    return a.argmin(axis=axis)

def where (condition,a,b) :
   return GradientReflector.where(condition=condition,a=a,b=b)

def sin(x) :
    if not isinstance(x,Tensor) : 
        x = Tensor(x)
    return x.sin()

def cos(x) :
    if not isinstance(x,Tensor) :
        x = Tensor(x)
    
    return x.cos()


from setuptools import setup ,find_packages

setup(
    name='littlelearn',
    version='1.0.7',
    description='machine learning ecosystem',
    long_description=open("README.md", encoding="utf-8").read(),
    author='Candra Alpin Gunawan',
    author_email='hinamatsuriairin@gmail.com',
    url='https://github.com/Airinchan818/LittleLearn',
    packages=find_packages(),
    install_requires=[

        'jax',
        'numpy',
        'tqdm',
        'pandas',
        'matplotlib',

    ],
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='deep learning,artificial intellengence,engine grad,ai,framework,api,machine learning',
    python_requires='>=3.7',

)



from .temp_optimizer import Optimizer
from .temp_optimizer import Adam
from .temp_optimizer import Adamax
from .temp_optimizer import AdamW
from .temp_optimizer import RMSProp
from .temp_optimizer import Lion



from .Activations import Relu
from .Activations import Leaky_Relu
from .Activations import Softmax
from .Activations import Sigmoid
from .Activations import Relu6 
from .Activations import LogSigmoid
from .Activations import LogSoftmax
from .Activations import HardShrink
from .Activations import HardSigmoid
from .Activations import HardTanh
from .Activations import Tanh
from .Activations import Swish
from .Activations import Selu
from .Activations import Celu
from .Activations import Elu 
from .Activations import Gelu
from .Activations import Softplus
from .Activations import Softsign
from .Activations import relu 
from .Activations import relu6
from .Activations import leaky_relu
from .Activations import log_sigmoid
from .Activations import log_softmax
from .Activations import sigmoid
from .Activations import softmax
from .Activations import tanh
from .Activations import hard_max
from .Activations import hard_shrink
from .Activations import hard_sigmoid
from .Activations import hard_tanh
from .Activations import swish
from .Activations import celu
from .Activations import selu
from .Activations import elu 
from .Activations import gelu
from .Activations import softplus
from .Activations import softsign
